@@ -59,6 +59,22 @@
 #include <unistd.h>
 #endif
 
+#if AP_ENABLE_EXCEPTION_HOOK
+APR_HOOK_STRUCT(
+    APR_HOOK_LINK(fatal_exception)
+    APR_HOOK_LINK(monitor)
+)
+AP_IMPLEMENT_HOOK_RUN_ALL(int, fatal_exception,
+                          (ap_exception_info_t *ei), (ei), OK, DECLINED)
+#else
+APR_HOOK_STRUCT(
+    APR_HOOK_LINK(monitor)
+)
+#endif
+AP_IMPLEMENT_HOOK_RUN_ALL(int, monitor,
+                          (apr_pool_t *p), (p), OK, DECLINED)
+
+
 #ifdef AP_MPM_WANT_RECLAIM_CHILD_PROCESSES
 
 typedef enum {DO_NOTHING, SEND_SIGTERM, SEND_SIGKILL, GIVEUP} action_t;
@@ -110,11 +126,6 @@ static int reclaim_one_pid(pid_t pid, action_t action)
     apr_proc_t proc;
     apr_status_t waitret;
 
-    /* Ensure pid sanity. */
-    if (pid < 1) {
-        return 1;
-    }        
-
     proc.pid = pid;
     waitret = apr_proc_wait(&proc, NULL, NULL, APR_NOWAIT);
     if (waitret != APR_CHILD_NOTDONE) {
@@ -124,7 +135,7 @@ static int reclaim_one_pid(pid_t pid, action_t action)
     switch(action) {
     case DO_NOTHING:
         break;
-        
+
     case SEND_SIGTERM:
         /* ok, now it's being annoying */
         ap_log_error(APLOG_MARK, APLOG_WARNING,
@@ -135,7 +146,7 @@ static int reclaim_one_pid(pid_t pid, action_t action)
                      pid);
         kill(pid, SIGTERM);
         break;
-        
+
     case SEND_SIGKILL:
         ap_log_error(APLOG_MARK, APLOG_ERR,
                      0, ap_server_conf,
@@ -155,7 +166,7 @@ static int reclaim_one_pid(pid_t pid, action_t action)
         kill_thread(pid);
 #endif
         break;
-                
+
     case GIVEUP:
         /* gave it our best shot, but alas...  If this really
          * is a child we are trying to kill and it really hasn't
@@ -170,7 +181,7 @@ static int reclaim_one_pid(pid_t pid, action_t action)
                      pid);
         break;
     }
-    
+
     return 0;
 }
 
@@ -242,7 +253,7 @@ void ap_reclaim_child_processes(int terminate)
                 ++not_dead_yet;
             }
         }
- 
+
         cur_extra = extras;
         while (cur_extra) {
             extra_process_t *next = cur_extra->next;
@@ -255,14 +266,45 @@ void ap_reclaim_child_processes(int terminate)
             }
             cur_extra = next;
         }
-
 #if APR_HAS_OTHER_CHILD
-        apr_proc_other_child_check();
+        apr_proc_other_child_refresh_all(APR_OC_REASON_RESTART);
 #endif
 
     } while (not_dead_yet > 0 &&
              action_table[cur_action].action != GIVEUP);
 }
+
+void ap_relieve_child_processes(void)
+{
+    int i;
+    extra_process_t *cur_extra;
+    int max_daemons;
+
+    ap_mpm_query(AP_MPMQ_MAX_DAEMON_USED, &max_daemons);
+
+    /* now see who is done */
+    for (i = 0; i < max_daemons; ++i) {
+        pid_t pid = MPM_CHILD_PID(i);
+
+        if (pid == 0) {
+            continue; /* not every scoreboard entry is in use */
+        }
+
+        if (reclaim_one_pid(pid, DO_NOTHING)) {
+            MPM_NOTE_CHILD_KILLED(i);
+        }
+    }
+
+    cur_extra = extras;
+    while (cur_extra) {
+        extra_process_t *next = cur_extra->next;
+
+        if (reclaim_one_pid(cur_extra->pid, DO_NOTHING)) {
+            AP_DEBUG_ASSERT(1 == ap_unregister_extra_mpm_process(cur_extra->pid));
+        }
+        cur_extra = next;
+    }
+}
 #endif /* AP_MPM_WANT_RECLAIM_CHILD_PROCESSES */
 
 #ifdef AP_MPM_WANT_WAIT_OR_TIMEOUT
@@ -281,6 +323,7 @@ void ap_wait_or_timeout(apr_exit_why_e *status, int *exitcode, apr_proc_t *ret,
     ++wait_or_timeout_counter;
     if (wait_or_timeout_counter == INTERVAL_OF_WRITABLE_PROBES) {
         wait_or_timeout_counter = 0;
+        ap_run_monitor(p);
     }
 
     rv = apr_proc_wait_all_procs(ret, exitcode, status, APR_NOWAIT, p);
@@ -316,7 +359,7 @@ int ap_process_child_status(apr_proc_t *pid, apr_exit_why_e why, int status)
      * check for bad rc from us and exit, running any
      * appropriate cleanups.
      *
-     * If the child died due to a resource shortage, 
+     * If the child died due to a resource shortage,
      * the parent should limit the rate of forking
      */
     if (APR_PROC_CHECK_EXIT(why)) {
@@ -471,8 +514,8 @@ AP_DECLARE(apr_status_t) ap_mpm_pod_open(apr_pool_t *p, ap_pod_t **pod)
     (*pod)->p = p;
 
     /* close these before exec. */
-    apr_file_unset_inherit((*pod)->pod_in);
-    apr_file_unset_inherit((*pod)->pod_out);
+    apr_file_inherit_unset((*pod)->pod_in);
+    apr_file_inherit_unset((*pod)->pod_out);
 
     return APR_SUCCESS;
 }
@@ -535,9 +578,11 @@ static apr_status_t pod_signal_internal(ap_pod_t *pod)
  */
 static apr_status_t dummy_connection(ap_pod_t *pod)
 {
+    char *srequest;
     apr_status_t rv;
     apr_socket_t *sock;
     apr_pool_t *p;
+    apr_size_t len;
 
     /* create a temporary pool for the socket.  pconf stays around too long */
     rv = apr_pool_create(&p, pod->p);
@@ -545,10 +590,12 @@ static apr_status_t dummy_connection(ap_pod_t *pod)
         return rv;
     }
 
-    rv = apr_socket_create(&sock, ap_listeners->bind_addr->family, SOCK_STREAM, p);
+    rv = apr_socket_create(&sock, ap_listeners->bind_addr->family,
+                           SOCK_STREAM, 0, p);
     if (rv != APR_SUCCESS) {
         ap_log_error(APLOG_MARK, APLOG_WARNING, rv, ap_server_conf,
                      "get socket to connect to listener");
+        apr_pool_destroy(p);
         return rv;
     }
 
@@ -563,10 +610,11 @@ static apr_status_t dummy_connection(ap_pod_t *pod)
         ap_log_error(APLOG_MARK, APLOG_WARNING, rv, ap_server_conf,
                      "set timeout on socket to connect to listener");
         apr_socket_close(sock);
+        apr_pool_destroy(p);
         return rv;
     }
 
-    rv = apr_connect(sock, ap_listeners->bind_addr);
+    rv = apr_socket_connect(sock, ap_listeners->bind_addr);
     if (rv != APR_SUCCESS) {
         int log_level = APLOG_WARNING;
 
@@ -582,6 +630,24 @@ static apr_status_t dummy_connection(ap_pod_t *pod)
                      "connect to listener on %pI", ap_listeners->bind_addr);
     }
 
+    /* Create the request string. We include a User-Agent so that
+     * adminstrators can track down the cause of the odd-looking
+     * requests in their logs.
+     */
+    srequest = apr_pstrcat(p, "GET / HTTP/1.0\r\nUser-Agent: ",
+                           ap_get_server_version(),
+                           " (internal dummy connection)\r\n\r\n", NULL);
+
+    /* Since some operating systems support buffering of data or entire
+     * requests in the kernel, we send a simple request, to make sure
+     * the server pops out of a blocking accept().
+     */
+    /* XXX: This is HTTP specific. We should look at the Protocol for each
+     * listener, and send the correct type of request to trigger any Accept
+     * Filters.
+     */
+    len = strlen(srequest);
+    apr_socket_send(sock, srequest, &len);
     apr_socket_close(sock);
     apr_pool_destroy(p);
 
@@ -709,7 +775,7 @@ const char *ap_mpm_set_coredumpdir(cmd_parms *cmd, void *dummy,
 
     fname = ap_server_root_relative(cmd->pool, arg);
     if (!fname) {
-        return apr_pstrcat(cmd->pool, "Invalid CoreDumpDirectory path ", 
+        return apr_pstrcat(cmd->pool, "Invalid CoreDumpDirectory path ",
                            arg, NULL);
     }
     if ((rv = apr_stat(&finfo, fname, APR_FINFO_TYPE, cmd->pool)) != APR_SUCCESS) {
@@ -726,6 +792,21 @@ const char *ap_mpm_set_coredumpdir(cmd_parms *cmd, void *dummy,
 }
 #endif
 
+#ifdef AP_MPM_WANT_SET_GRACEFUL_SHUTDOWN
+int ap_graceful_shutdown_timeout = 0;
+
+const char * ap_mpm_set_graceful_shutdown(cmd_parms *cmd, void *dummy,
+                                          const char *arg)
+{
+    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err != NULL) {
+        return err;
+    }
+    ap_graceful_shutdown_timeout = atoi(arg);
+    return NULL;
+}
+#endif
+
 #ifdef AP_MPM_WANT_SET_ACCEPT_LOCK_MECH
 apr_lockmech_e ap_accept_lock_mech = APR_LOCK_DEFAULT;
 
@@ -821,7 +902,7 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
     int running = 0;
     int have_pid_file = 0;
     const char *status;
-    
+
     *exit_status = 0;
 
     rv = ap_read_pid(pconf, ap_pid_fname, &otherpid);
@@ -829,6 +910,8 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
         if (rv != APR_ENOENT) {
             ap_log_error(APLOG_MARK, APLOG_STARTUP, rv, NULL,
                          "Error retrieving pid file %s", ap_pid_fname);
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
+                         "Remove it before continuing if it is corrupted.");
             *exit_status = 1;
             return 1;
         }
@@ -838,7 +921,7 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
         have_pid_file = 1;
         if (kill(otherpid, 0) == 0) {
             running = 1;
-            status = apr_psprintf(pconf, 
+            status = apr_psprintf(pconf,
                                   "httpd (pid %" APR_PID_T_FMT ") already "
                                   "running", otherpid);
         }
@@ -881,11 +964,25 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
             printf("httpd not running, trying to start\n");
         }
         else {
-            *exit_status = send_signal(otherpid, SIGUSR1);
+            *exit_status = send_signal(otherpid, AP_SIG_GRACEFUL);
             return 1;
         }
     }
 
+    if (!strcmp(dash_k_arg, "graceful-stop")) {
+#ifdef AP_MPM_WANT_SET_GRACEFUL_SHUTDOWN
+        if (!running) {
+            printf("%s\n", status);
+        }
+        else {
+            *exit_status = send_signal(otherpid, AP_SIG_GRACEFUL_STOP);
+        }
+#else
+        printf("httpd MPM \"" MPM_NAME "\" does not support graceful-stop\n");
+#endif
+        return 1;
+    }
+
     return 0;
 }
 
@@ -913,7 +1010,8 @@ void ap_mpm_rewrite_args(process_rec *process)
         case 'k':
             if (!dash_k_arg) {
                 if (!strcmp(optarg, "start") || !strcmp(optarg, "stop") ||
-                    !strcmp(optarg, "restart") || !strcmp(optarg, "graceful")) {
+                    !strcmp(optarg, "restart") || !strcmp(optarg, "graceful") ||
+                    !strcmp(optarg, "graceful-stop")) {
                     dash_k_arg = optarg;
                     break;
                 }
@@ -951,17 +1049,17 @@ void ap_mpm_rewrite_args(process_rec *process)
 apr_uint32_t ap_max_mem_free = APR_ALLOCATOR_MAX_FREE_UNLIMITED;
 
 const char *ap_mpm_set_max_mem_free(cmd_parms *cmd, void *dummy,
-	                            const char *arg)
+                                    const char *arg)
 {
     long value;
     const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
     if (err != NULL) {
         return err;
     }
-    
+
     value = strtol(arg, NULL, 0);
     if (value < 0 || errno == ERANGE)
-        return apr_pstrcat(cmd->pool, "Invalid MaxMemFree value: ", 
+        return apr_pstrcat(cmd->pool, "Invalid MaxMemFree value: ",
                            arg, NULL);
 
     ap_max_mem_free = (apr_uint32_t)value * 1024;
@@ -971,6 +1069,30 @@ const char *ap_mpm_set_max_mem_free(cmd_parms *cmd, void *dummy,
 
 #endif /* AP_MPM_WANT_SET_MAX_MEM_FREE */
 
+#ifdef AP_MPM_WANT_SET_STACKSIZE
+apr_size_t ap_thread_stacksize = 0; /* use system default */
+
+const char *ap_mpm_set_thread_stacksize(cmd_parms *cmd, void *dummy,
+                                        const char *arg)
+{
+    long value;
+    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
+    if (err != NULL) {
+        return err;
+    }
+
+    value = strtol(arg, NULL, 0);
+    if (value < 0 || errno == ERANGE)
+        return apr_pstrcat(cmd->pool, "Invalid ThreadStackSize value: ",
+                           arg, NULL);
+
+    ap_thread_stacksize = (apr_size_t)value;
+
+    return NULL;
+}
+
+#endif /* AP_MPM_WANT_SET_STACKSIZE */
+
 #ifdef AP_MPM_WANT_FATAL_SIGNAL_HANDLER
 
 static pid_t parent_pid, my_pid;
@@ -989,7 +1111,7 @@ const char *ap_mpm_set_exception_hook(cmd_parms *cmd, void *dummy,
     }
 
     if (cmd->server->is_virtual) {
-	return "EnableExceptionHook directive not allowed in <VirtualHost>";
+        return "EnableExceptionHook directive not allowed in <VirtualHost>";
     }
 
     if (strcasecmp(arg, "on") == 0) {
@@ -1005,19 +1127,12 @@ const char *ap_mpm_set_exception_hook(cmd_parms *cmd, void *dummy,
     return NULL;
 }
 
-APR_HOOK_STRUCT(
-    APR_HOOK_LINK(fatal_exception)
-)
-
-AP_IMPLEMENT_HOOK_RUN_ALL(int, fatal_exception,
-                          (ap_exception_info_t *ei), (ei), OK, DECLINED)
-
 static void run_fatal_exception_hook(int sig)
 {
     ap_exception_info_t ei = {0};
 
     if (exception_hook_enabled &&
-        geteuid() != 0 && 
+        geteuid() != 0 &&
         my_pid != parent_pid) {
         ei.sig = sig;
         ei.pid = my_pid;
@@ -1074,7 +1189,7 @@ apr_status_t ap_fatal_signal_setup(server_rec *s, apr_pool_t *in_pconf)
     struct sigaction sa;
 
     sigemptyset(&sa.sa_mask);
-    
+
 #if defined(SA_ONESHOT)
     sa.sa_flags = SA_ONESHOT;
 #elif defined(SA_RESETHAND)
@@ -1104,7 +1219,7 @@ apr_status_t ap_fatal_signal_setup(server_rec *s, apr_pool_t *in_pconf)
 #endif
 
 #else /* NO_USE_SIGACTION */
-    
+
     apr_signal(SIGSEGV, sig_coredump);
 #ifdef SIGBUS
     apr_signal(SIGBUS, sig_coredump);
